# c2cj设计文档

## 1. C语言转换为仓颉语言语法的主要类别

* 函数声明
* 结构体定义
* 全局变量(extern)
* 枚举（enum）
* 别名（typedef）



### 函数声明

 举例模板：

```c
// c language
void *malloc(size_t size);
// cj language
foreign func malloc(size: UIntNative): CPointer(Unit)
```



### 结构体定义

```c
// c language
struct info {
	int a;
	char b;
    struct todo {
		void c;
	};
};

// cj language 
record info {
    var a: Int64;
    var b: UInt8;
    record todo {
        var c: Unit;
    };
};
```



## 2. 文件解析

​    先从文件中解析处`注释`，` 结构体`，`  函数`等特性字段后， 根据对应结构特性进行解析转码成仓颉语法。



- #### 解析思路

```c
/* *******************
 * c 2 cj parser tools
 * 2022-4
 * hw
 * *******************/
void *func(size_t size, int a, void b, const char *);

//infomation for school
struct info {
	int a;
	char *b;
    struct todo {
		void c;
	};
	struct info *next;
};
```

1. 注释解析:   由于在C中如果带有`//`或`/* */`的语句就视为是注释语句，所以注释的优先级较高，我们首先判断读取的语句是否是注释

   ```c
   /* *******************
    * c 2 cj parser tools
    * 2022-4
    * hw
    * *******************/
   //infomation for school
   ```

   - 代码逻辑实现设计

   ```c
   for (c in strRead)      //strRead是readAllText()读到的所有内容
       str += c
       if (cc == '/' && c == '*')
           notesflag = 1
       else if (cc == '/' && c == '/')
           notesflag = 2
       if (notesflag == 1 && (c == '\n' && cc == '/'))      //读取"/* ... */，将str写到Native.cj"
           str is "* ... */"
           notesflag = 0
           str = ""
       else if (notesFlag == 2 && c == '\n')
           str is "//..."                                   //读取"//..."
           notesflag = 0
           str = ""
       cc = c
   ```

2. 结构体解析：如果在接收的字符中不带有注释标识符并且有`{`，就证明开始接收的字符串是一个结构体，以`;`和`{ }`对应数量相同作为结束

   ```c
   struct info {
   	int a;
   	char *b;
       struct todo {
   		void c;
   	};
   	struct info *next;
   };
   ```

   - 代码逻辑实现设计

   ```c
   for (c in strRead)
   	str += c
       if (c == '{' && notesflag == 0)
           structflag = 1
       if (structflag == 1 && c == ';' && lcount == rcount)
           str is "struct info { };"
           str = ""
   ```

3. 函数解析：如果截取的字符串既不是注释也不是结构体，如果末尾结尾是`);`形式的则证明是函数

   ```c
   void *func(size_t size, int a, void b, const char *);
   ```

   - 代码逻辑实现设计

   ```c
   for (c in strRead)
       str += c
       if (notesflag == 0 && structflag == 0)
           funcflag = 1
       if (funcflag == 1 && c == ';' && cc == ')')
           str is "func();"
           str = ""
       else if (funcflag == 1 && c == '\n' && cc != ';')
           str is "不需要解析的部分，例如#define ...  #include ..."
           str = ""
    	cc = c
   ```



## 3. 转码解析

`c lang` 与 `cj lang`数据类型映射:

```c
let cType2cjType = HashMap<String, String>([("void", "Unit"), ("bool", "Bool"), ("char", "UInt8"), ("int", "Int64"), ("int8_t", "Int8"), ("uint8_t", "UInt8"), ("int16_t", "Int16"), ("uint16_t", "UInt16"), ("int32_t", "Int32"), ("uint32_t", "UInt32"), ("int64_t", "Int64"), ("uint64_t", "UInt64"), ("", "IntNative"), ("", "UintNative"), ("float", "Float32"), ("double", "Float64"), ("*", "CPointer"), ("char *", "CString"), ("uint8_t*", "CString"), ("int8_t*", "CString")])
```



### 函数解析

```c
void *func(size_t size, int a, void b, const char *c...);
```

- #### 解析思路：

  ​	1. 根据函数的组成标识符将函数分为两部分，（函数类型+函数名）和参数             //只解析有一对`(``)`的函数，多括号情况暂时不能处理

  ```c
  void *func   --1
  (size_t size, int a, void b, char *c...)  --2
  ```

  2. 如果函数类型是标准C类型，则在上面`HashMap`中找到对应CJ中类型，如果带有指针`*`，则转出对应的`CPointer<>`形式

  ```c
  void <=> Unit, int <=> Int64, * <=> CPointer<>...
  ```

  - 代码逻辑实现设计

  ```c
  var index = match (' ')
  var cfunctype = funcstr.substring(0, index) //此处省略句首含有空格情况
  var cjfunctype = cType2cjType[cfunctype]    //cType2cjType为上面HashMap方法
  ```

  3. 截取参数部分，首先根据`(`和`)`进行查找，由于函数指针同样用括号包含，所以在查找标识符是采用最后一对`(``)`为准

  ```c
  (size_t size, ...)
  ```

  - 代码逻辑实现设计

  ```c
  var rIndex = match(funcstr.lastIndexOf(')'))
  var lIndex = match(funcstr.lastIndexOf('('))
  var cparamstr = funcstr.substring(lIndex + 1, rIndex - lIndex - 1)
  ```

  4. 参数部分解析，参数以`,`分隔，逐个解析逗号之间的参数

  ```c
  (size_t size, int a, void b, const char *c,...)
  ```

  - 代码逻辑实现设计

  ```c
  do {
  	var index = match(',')
  	var curstr = cparamstr.substring(0, index)  //句首空格已处理
  	childparamparser(curstr)
  	cparamstr = cparamstr.substring(index + 1)
  } while(index != -1)
  ```

  预期结果

  ```c
  size_t size
  int a
  void b
  const char *c
  ...
  ```

  - 单个参数解析，指针参数，结构体参数，标准参数...

  ```c
  size_t size, int a, void b, const char *c,...
  ```

  - 代码逻辑实现设计

  ```c
  cIndex = match(' ')                    //判断是否有空格
  pIndex = match('*')                    //判断是否有指针
  sIndex = match("struct")               //判断是否是结构体
  paramstr = paramstr.replace("struct", "")  //如果有struct，去掉， paramstr为单个参数字符串
  if pIndex != -1
  	paramType = paramstr.substring(0, pIndex)    //参数类型截取到*，之后用replace()去空格
      paramValue = paramstr.substring(pIndex + 1) 
  else if cIndex != -1
      paramType = paramstr.substring(0, cIndex)    //参数类型截取到空格
      paramValue = ...
  else
      "don't handle"
  cjparamtype = cType2cjType[paramType]            //如果是标准类型，找到对应的仓颉标准类型
  ```

  5. 参数带`const`的直接去掉

  ```c
  const char c   =>   char c
  ```

  6. 复杂的函数暂不解析

  ```c
  int (*func)(int *p, int (*f)(int *));
  int func(char **c)
  ```

### 结构体解析

```c
struct info {
	int a;
	char *b;
    struct todo {
		void c;        /* c is ... */
	};
	struct info *next;
};
```

- ### 解析思路

  1. 结构体主要由“结构名”，花括号`{ };`，成员变量（类型+变量名）组成，首先根据结构体特征`{ }`来截取一个完整的结构体
  2. 结构体可以分为`typedef struct{ }A;`，`struct B{ };`，别名+`{ };`等
  3. 根据左右`{ }`把结构体分为两部分，成员变量和结构名

  ```c
  struct info { };          ----1
  int a;                    ----2
  char *b;
  struct todo {            
  		void c;   
  	};
  struct info *next;
  ```

  - 代码逻辑实现设计

  ```  c
  rIndex = match(structstr.lastIndexOf('}'))
  sIndex = match(structstr.lastIndexOf(';'))
  lIndex = match(structstr.IndexOf('{'))     
  structmemstr = structstr.substring(lIndex + 1, rIndex - lIndex -1)    //成员部分
  structmanestr = structstr.substring(0, lIndex)                        //结构名
  structmanestr = structmanestr.replace("typedef/struct", "")
  for(c in structmanestr)
      if c.isAsciiLetter()         //判断是否有结构名
          structmanestr
      else
      	structmanestr = structstr.substring(rIndex + 1, sIndex - rIndex - 1)    //前面没有结构名，就去后面`}`和`;`之间找
  ```

  4. 成员变量解析，根据`;`截取每个子成员变量，如果子成员变量是结构体，返回步骤3进行结构体解析

  ```c
  int a;                    ----普通成员
  char *b;
  struct todo {             ----结构体  
  		void c;
  	};
  struct info *next;        ----链表
  ```

  - 代码逻辑实现设计

  ```c
  structmemfunc()
      for (c in structmemstr)
          childmemstr += c
          if (c == '{')
              lcount++                      //后面判断是否是结构体
          else if (c == '}')
              rcount++
          if (c == ';' && lcount == rcount)
              childmemfunc(childmemstr);               //调用成员解析函数
  
  childmemfunc()
     lIndex = match(structstr.IndexOf('{'))           
     if (lIndex != -1)                            //如果是结构体，返回调用structmemfunc函数，return
         structmemfunc()      
         return
     else
         childmemstr = childmemstr.replace("struct", "")     //去掉struct或者const等不解析的标识符
         index = match(structstr.IndexOf('*'/' '))          //判断是否有指针，根据指针或者空格解析出类型和变量
  ```

  5. 如果结构体中含有`/* ... */`，根据注释解析部分进行解析

  ```c
  struct todo {               
  		void c;     /* c is ... */
  };
  ```

  6. 结构体含有函数，待解析...

## 4. 特殊类型语法字段解析

- 函数指针和指针函数仓颉语法怎样区分

  - 函数指针

  ```c
  	int(*p)(int, int);
  ```

  - 指针函数

  ```c
  	int *p(int, int);
  ```

- 多指针变量怎样表示

```c
int a(char **c);
```

- 参数不带变量名需要怎么显示

```c
int func(int, void *);
```

- 数据类型别名和结构体别名怎样区分

```c
size_t -> UIntNative         ST A
```

- 结构体嵌有union联合

```c
struct FD {
	int a;
	union {
      void b;  
    };
};
```

- 头文件里含有宏条件编译

```c
#ifdef __cplusplus
#if __cplusplus
extern "C" {
#endif /* __cplusplus */
#endif /* __cplusplus */
---------------------------

---------------------------
#ifdef __cplusplus
#if __cplusplus
}
#endif /* __cplusplus */
#endif /* __cplusplus */
```

- 参数带const

```c
const int a -> var a: Int64
```
